#include "soundcloud.h"
#include "json.h"
#include "utils.h"
#include <QRegExp>
#include <QNetworkAccessManager>
#include <QNetworkRequest>
#include <QNetworkReply>
#include <QStringList>
#include <QDateTime>
#include <QUrl>
#include <QDebug>
#ifdef QML_USER_INTERFACE
#include <QDeclarativeEngine>
#endif

const QString CLIENT_ID("aada80a82c91fb42d0bb1cb321b59034");
const QString CLIENT_SECRET("dcc9c80ce05614d4988cebb82ddde152");
const QString REDIRECT_URI("http://www.musikloud.co.uk");

SoundCloud* soundCloudInstance = 0;

using namespace QtJson;

SoundCloud::SoundCloud(QObject *parent) :
    QObject(parent),
    m_nam(0),
    m_busy(false),
    m_cancelled(false)
{
    if (!soundCloudInstance) {
        soundCloudInstance = this;
    }
}

SoundCloud::~SoundCloud() {}

SoundCloud* SoundCloud::instance() {
    return soundCloudInstance;
}

void SoundCloud::setBusy(bool isBusy, const QString &message, int numberOfOperations) {
    if (isBusy != this->busy()) {
        m_busy = isBusy;

        if (isBusy) {
            this->setCancelled(false);
            emit busy(message, numberOfOperations);
        }
        else if (!this->cancelled()) {
            emit busyProgressChanged(numberOfOperations);
        }

        emit busyChanged(isBusy);
    }
}

void SoundCloud::cancelCurrentOperation() {
    this->setCancelled(true);
    this->setBusy(false);
    this->disconnect(this, SIGNAL(postSuccessful(QString)), this, 0);
    this->disconnect(this, SIGNAL(postFailed(QString)), this, 0);
    this->disconnect(this, SIGNAL(gotResourceId(QString,Queries::QueryType)), 0, 0);
#ifdef QML_USER_INTERFACE
    this->disconnect(this, SIGNAL(gotTrack(TrackItem*)), 0, 0);
    this->disconnect(this, SIGNAL(gotPlaylist(PlaylistItem*)), 0, 0)
        #else
    this->disconnect(this, SIGNAL(gotTrack(QSharedPointer<TrackItem>)), 0, 0);
    this->disconnect(this, SIGNAL(gotPlaylist(QSharedPointer<PlaylistItem>)), 0, 0);
#endif
    emit currentOperationCancelled();
}

QUrl SoundCloud::authUrl() {
    QUrl url("https://soundcloud.com/connect");
    url.addQueryItem("client_id", CLIENT_ID);
    url.addQueryItem("response_type", "code");
    url.addQueryItem("redirect_uri", REDIRECT_URI);
    url.addQueryItem("scope", "non-expiring");
    url.addQueryItem("display", "popup");

    return url;
}

QNetworkReply* SoundCloud::createReply(QUrl url, int offset) {
    if (url.scheme() == "https") {
        if (this->userSignedIn()) {
            url.addQueryItem("oauth_token", this->accessToken());
        }
        else {
            url.setScheme("http");
            url.addQueryItem("client_id", CLIENT_ID);
        }
    }
    else {
        url.addQueryItem("client_id", CLIENT_ID);
    }

    if (offset) {
        url.addQueryItem("offset", QString::number(offset));
    }

    qDebug() << url;

    QNetworkRequest request(url);
    request.setRawHeader("User-Agent", QString("MusiKloud/%1 (Qt)").arg(Utils::versionNumberString()).toUtf8());

    return this->networkAccessManager()->get(request);
}

QNetworkReply* SoundCloud::createSearchReply(int queryType, const QString &query, int order, int offset) {
    QString qs(query);
    qs.replace(' ', '+');

    QUrl url;

    switch (queryType) {
    case Queries::Tracks:
        url = TRACKS_BASE_FEED;
        url.addQueryItem("filter", "streamable");
        break;
    case Queries::Playlists:
        url = PLAYLISTS_BASE_FEED;
        break;
    case Queries::Users:
        url = USERS_BASE_FEED;
        break;
    case Queries::Groups:
        url = GROUPS_BASE_FEED;
        break;
    default:
        qWarning() << "SoundCloud::createSearchReply(): Invalid query type.";
        break;
    }

    switch (order) {
    case Queries::Date:
        url.addQueryItem("order", "created_at");
        break;
    default:
        url.addQueryItem("order", "hotness");
        break;
    }

    url.addQueryItem("q", "'" + query + "'|" + qs);
    url.addQueryItem("client_id", CLIENT_ID);

    if (offset) {
        url.addQueryItem("offset", QString::number(offset));
    }

    qDebug() << url;

    QNetworkRequest request(url);
    request.setRawHeader("User-Agent", QString("MusiKloud/%1 (Qt)").arg(Utils::versionNumberString()).toUtf8());

    return this->networkAccessManager()->get(request);
}

QUrl SoundCloud::getStreamUrl(QUrl url) {
    if (url.scheme() == "https") {
        url.setScheme("http");
    }

    url.addQueryItem("client_id", CLIENT_ID);

    return url;
}

QUrl SoundCloud::getDownloadUrl(QUrl url) {
    if (this->userSignedIn()) {
        url.addQueryItem("oauth_token", this->accessToken());
    }
    else {
        url.addQueryItem("client_id", CLIENT_ID);
    }

    return url;
}

void SoundCloud::resolve(const QString &urlToResolve) {
    this->setBusy(true, tr("Resolving URL"));
    QUrl url("https://api.soundcloud.com/resolve");
    url.addQueryItem("url", urlToResolve);
    url.addQueryItem("client_id", CLIENT_ID);
    QNetworkRequest request(url);
    request.setRawHeader("User-Agent", QString("MusiKloud/%1 (Qt)").arg(Utils::versionNumberString()).toUtf8());
    QNetworkReply *reply = this->networkAccessManager()->get(request);
    this->connect(reply, SIGNAL(finished()), this, SLOT(checkResolvedUrl()));
    this->connect(this, SIGNAL(currentOperationCancelled()), reply, SLOT(deleteLater()));
}

void SoundCloud::checkResolvedUrl() {
    this->setBusy(false);
    QNetworkReply* reply = qobject_cast<QNetworkReply*>(this->sender());

    if (!reply) {
        emit error(tr("Network error"));
        return;
    }

    QString resource = reply->header(QNetworkRequest::LocationHeader).toString();
    QString id = resource.section('/', -1).section('?', 0, 0);

    if (!id.isEmpty()) {
        QString resourceType = resource.section('/', -2, -2);

        if (resourceType == "tracks") {
            emit gotResourceId(id, Queries::Tracks);
        }
        else if (resourceType == "playlists") {
            emit gotResourceId(id, Queries::Playlists);
        }
        else if (resourceType == "users") {
            emit gotResourceId(id, Queries::Users);
        }
        else {
            emit error(tr("Unable to resolve URL"));
        }
    }
    else {
        emit error(tr("Unable to resolve URL"));
    }

    this->disconnect(this, SIGNAL(gotResourceId(QString,Queries::QueryType)), 0, 0);

    reply->deleteLater();
}

void SoundCloud::getFullTrack(const QString &id) {
    this->setBusy(true, tr("Retrieving track details"));

    if (id.contains("soundcloud.com")) {
        this->resolve(id);
        this->connect(this, SIGNAL(gotResourceId(QString,Queries::QueryType)), this, SLOT(getFullTrack(QString)));
        return;
    }

    QUrl url;

    if (this->userSignedIn()) {
        url.setUrl("https://api.soundcloud.com/tracks/" + id + ".json");
        url.addQueryItem("oauth_token", this->accessToken());
    }
    else {
        url.setUrl("http://api.soundcloud.com/tracks/" + id + ".json");
        url.addQueryItem("client_id", CLIENT_ID);
    }

    QNetworkRequest request(url);
    request.setRawHeader("User-Agent", QString("MusiKloud/%1 (Qt)").arg(Utils::versionNumberString()).toUtf8());
    QNetworkReply *reply = this->networkAccessManager()->get(request);
    this->connect(reply, SIGNAL(finished()), this, SLOT(checkTrack()));
    this->connect(this, SIGNAL(currentOperationCancelled()), reply, SLOT(deleteLater()));
}

void SoundCloud::checkTrack() {
    this->setBusy(false);
    QNetworkReply* reply = qobject_cast<QNetworkReply*>(this->sender());

    if (!reply) {
        emit error(tr("Network error"));
        return;
    }

    QString response(reply->readAll());
    bool ok;
    QVariantMap result = Json::parse(response, ok).toMap();

    if (!ok) {
        int statusCode = reply->attribute(QNetworkRequest::HttpStatusCodeAttribute).toInt();

        if (statusCode != 200) {
            emit error(reply->attribute(QNetworkRequest::HttpReasonPhraseAttribute).toString());
        }
        else {
            emit error(tr("Cannot parse server response"));
        }
    }
    else {
        TrackItem *track = new TrackItem(result);
#ifdef QML_USER_INTERFACE
        QDeclarativeEngine::setObjectOwnership(track, QDeclarativeEngine::JavaScriptOwnership);
        emit gotTrack(track);
#else
        emit gotTrack(QSharedPointer<TrackItem>(track));
#endif
    }
#ifdef QML_USER_INTERFACE
    this->disconnect(this, SIGNAL(gotTrack(TrackItem*)), 0, 0);
#else
    this->disconnect(this, SIGNAL(gotTrack(QSharedPointer<TrackItem>)), 0, 0);
#endif
    reply->deleteLater();
}

void SoundCloud::getFullPlaylist(const QString &id) {
    this->setBusy(true, tr("Retrieving set details"));

    if (id.contains("soundcloud.com")) {
        this->resolve(id);
        this->connect(this, SIGNAL(gotResourceId(QString,Queries::QueryType)), this, SLOT(getFullPlaylist(QString)));
        return;
    }

    QUrl url;

    if (this->userSignedIn()) {
        url.setUrl("https://api.soundcloud.com/playlists/" + id + ".json");
        url.addQueryItem("oauth_token", this->accessToken());
    }
    else {
        url.setUrl("http://api.soundcloud.com/playlists/" + id + ".json");
        url.addQueryItem("client_id", CLIENT_ID);
    }

    QNetworkRequest request(url);
    request.setRawHeader("User-Agent", QString("MusiKloud/%1 (Qt)").arg(Utils::versionNumberString()).toUtf8());
    QNetworkReply *reply = this->networkAccessManager()->get(request);
    this->connect(reply, SIGNAL(finished()), this, SLOT(checkPlaylist()));
    this->connect(this, SIGNAL(currentOperationCancelled()), reply, SLOT(deleteLater()));
}

void SoundCloud::checkPlaylist() {
    this->setBusy(false);
    QNetworkReply* reply = qobject_cast<QNetworkReply*>(this->sender());

    if (!reply) {
        emit error(tr("Network error"));
        return;
    }

    QString response(reply->readAll());
    bool ok;
    QVariantMap result = Json::parse(response, ok).toMap();

    if (!ok) {
        int statusCode = reply->attribute(QNetworkRequest::HttpStatusCodeAttribute).toInt();

        if (statusCode != 200) {
            emit error(reply->attribute(QNetworkRequest::HttpReasonPhraseAttribute).toString());
        }
        else {
            emit error(tr("Cannot parse server response"));
        }
    }
    else {
        PlaylistItem *playlist = new PlaylistItem(result);
#ifdef QML_USER_INTERFACE
        QDeclarativeEngine::setObjectOwnership(playlist, QDeclarativeEngine::JavaScriptOwnership);
        emit gotPlaylist(playlist);
#else
        emit gotPlaylist(QSharedPointer<PlaylistItem>(playlist));
#endif
    }
#ifdef QML_USER_INTERFACE
    this->disconnect(this, SIGNAL(gotPlaylist(PlaylistItem*)), 0, 0);
#else
    this->disconnect(this, SIGNAL(gotPlaylist(QSharedPointer<PlaylistItem>)), 0, 0);
#endif
    reply->deleteLater();
}

void SoundCloud::getUserProfile(const QString &id) {
    QUrl url(QString("http://api.soundcloud.com/users/%1.json").arg(id));
    url.addQueryItem("client_id", CLIENT_ID);
    QNetworkRequest request(url);
    request.setRawHeader("User-Agent", QString("MusiKloud/%1 (Qt)").arg(Utils::versionNumberString()).toUtf8());
    QNetworkReply *reply = this->networkAccessManager()->get(request);
    this->connect(reply, SIGNAL(finished()), this, SLOT(checkUser()));
}

void SoundCloud::getCurrentUserProfile() {
    QUrl url("https://api.soundcloud.com/me.json");
    url.addQueryItem("oauth_token", this->accessToken());
    QNetworkRequest request(url);
    request.setRawHeader("User-Agent", QString("MusiKloud/%1 (Qt)").arg(Utils::versionNumberString()).toUtf8());
    QNetworkReply *reply = this->networkAccessManager()->get(request);
    this->connect(reply, SIGNAL(finished()), this, SLOT(checkCurrentUser()));
}

void SoundCloud::checkUser() {
    QNetworkReply* reply = qobject_cast<QNetworkReply*>(this->sender());

    if (!reply) {
        emit error(tr("Network error"));
        return;
    }

    QString response(reply->readAll());
    bool ok;
    QVariantMap result = Json::parse(response, ok).toMap();

    if (!ok) {
        emit error(tr("Cannot parse server response"));
        return;
    }

    int statusCode = reply->attribute(QNetworkRequest::HttpStatusCodeAttribute).toInt();

    if (statusCode != 200) {
        emit error(reply->attribute(QNetworkRequest::HttpReasonPhraseAttribute).toString());
    }
    else {
        UserItem *user = new UserItem(result);

        if (!user->followed()) {
            this->isUserFollowed(user->id());
        }
#ifdef QML_USER_INTERFACE
        emit gotUser(user);
#else
        emit gotUser(QSharedPointer<UserItem>(user));
#endif
    }

    reply->deleteLater();
}

void SoundCloud::checkCurrentUser() {
    QNetworkReply* reply = qobject_cast<QNetworkReply*>(this->sender());

    if (!reply) {
        emit error(tr("Network error"));
        return;
    }

    QString response(reply->readAll());
    bool ok;
    QVariantMap result = Json::parse(response, ok).toMap();

    if (!ok) {
        int statusCode = reply->attribute(QNetworkRequest::HttpStatusCodeAttribute).toInt();

        if (statusCode != 200) {
            emit error(reply->attribute(QNetworkRequest::HttpReasonPhraseAttribute).toString());

            if (statusCode == 401) {
                emit signedIn(QString());
            }
        }
        else {
            emit error(tr("Cannot parse server response"));
        }
    }

    UserItem *user = new UserItem(result);

    if (this->username().isEmpty()) {
        this->setUsername(user->username());
        emit gotUsername(user->username());
    }
#ifdef QML_USER_INTERFACE
    emit gotUser(user);
#else
    emit gotUser(QSharedPointer<UserItem>(user));
#endif
    reply->deleteLater();
}

void SoundCloud::isUserFollowed(const QString &id) {
    m_actionId = id;
    QUrl url(QString("https://api.soundcloud.com/me/followings/%1.json").arg(id));
    url.addQueryItem("oauth_token", this->accessToken());
    QNetworkRequest request(url);
    request.setRawHeader("User-Agent", QString("MusiKloud/%1 (Qt)").arg(Utils::versionNumberString()).toUtf8());
    QNetworkReply *reply = this->networkAccessManager()->get(request);
    this->connect(reply, SIGNAL(finished()), this, SLOT(checkIfUserFollowed()));
}

void SoundCloud::checkIfUserFollowed() {
    QNetworkReply* reply = qobject_cast<QNetworkReply*>(this->sender());

    if (!reply) {
        emit error(tr("Network error"));
        return;
    }

    emit followingChanged(m_actionId, reply->attribute(QNetworkRequest::HttpStatusCodeAttribute).toInt() != 404);
    reply->deleteLater();
}

void SoundCloud::trackIsFavourite(const QString &id) {
    m_actionId = id;
    QUrl url(QString("https://api.soundcloud.com/me/favorites/%1.json").arg(id));
    url.addQueryItem("oauth_token", this->accessToken());
    QNetworkRequest request(url);
    request.setRawHeader("User-Agent", QString("MusiKloud/%1 (Qt)").arg(Utils::versionNumberString()).toUtf8());
    QNetworkReply *reply = this->networkAccessManager()->get(request);
    this->connect(reply, SIGNAL(finished()), this, SLOT(checkIfFavourite()));
}

void SoundCloud::checkIfFavourite() {
    QNetworkReply* reply = qobject_cast<QNetworkReply*>(this->sender());

    if (!reply) {
        emit error(tr("Network error"));
        return;
    }

    emit favouriteChanged(m_actionId, reply->attribute(QNetworkRequest::HttpStatusCodeAttribute).toInt() != 404);
    reply->deleteLater();
}

void SoundCloud::signIn(const QUrl &response) {
    QByteArray authCode = response.encodedQueryItemValue("code");

    if (authCode.isEmpty()) {

        QString errorString = response.queryItemValue("error");

        if (errorString == "access_denied") {
            emit info(tr("You have denied access to your SoundCloud account"));
        }
        else if (errorString.isEmpty()){
            emit error(tr("No authorisation code received from SoundCloud"));
        }
        else {
            emit error(response.queryItemValue("error_description").replace('+', ' '));
        }

        return;
    }

    this->setBusy(true, tr("Signing in"));
    QNetworkRequest request(QUrl("https://api.soundcloud.com/oauth2/token"));
    QByteArray data(
                "client_id=" + CLIENT_ID.toUtf8()
                + "&client_secret=" + CLIENT_SECRET.toUtf8()
                + "&redirect_uri=" + REDIRECT_URI.toUtf8()
                + "&grant_type=authorization_code"
                + "&code=" + authCode
                );

    request.setHeader(QNetworkRequest::ContentTypeHeader, "application/x-www-form-urlencoded");
    request.setRawHeader("User-Agent", QString("MusiKloud/%1 (Qt)").arg(Utils::versionNumberString()).toUtf8());
    QNetworkReply *reply = this->networkAccessManager()->post(request, data);
    this->connect(reply, SIGNAL(finished()), this, SLOT(checkAccessToken()));
    this->connect(this, SIGNAL(currentOperationCancelled()), reply, SLOT(deleteLater()));
}

void SoundCloud::signIn(const QString &user, const QString &pass) {
    this->setBusy(true, tr("Signing in"));
    QNetworkRequest request(QUrl("https://api.soundcloud.com/oauth2/token"));
    QByteArray data(
                "client_id=" + CLIENT_ID.toUtf8()
                + "&client_secret=" + CLIENT_SECRET.toUtf8()
                + "&scope=non-expiring"
                + "&grant_type=password"
                + "&username=" + user.toUtf8()
                + "&password=" + pass.toUtf8()
                );

    request.setHeader(QNetworkRequest::ContentTypeHeader, "application/x-www-form-urlencoded");
    request.setRawHeader("User-Agent", QString("MusiKloud/%1 (Qt)").arg(Utils::versionNumberString()).toUtf8());
    QNetworkReply *reply = this->networkAccessManager()->post(request, data);
    this->connect(reply, SIGNAL(finished()), this, SLOT(checkAccessToken()));
    this->connect(this, SIGNAL(currentOperationCancelled()), reply, SLOT(deleteLater()));
}

void SoundCloud::checkAccessToken() {
    this->setBusy(false);
    QNetworkReply* reply = qobject_cast<QNetworkReply*>(this->sender());

    if (!reply) {
        emit error(tr("Network error"));
        return;
    }

    QString response(reply->readAll());
    bool ok;
    QVariantMap result = Json::parse(response, ok).toMap();

    if (!ok) {
        emit error(tr("Error parsing server response"));
    }
    else {
        QString token = result.value("access_token").toString();

        if (token.isEmpty()) {

            QString errorString = result.value("error").toString();

            if (!errorString.isEmpty()) {
                emit error(errorString);
            }
            else {
                emit error(tr("Error obtaining access token"));
            }
        }
        else {
            emit signedIn(token);
            emit alert(tr("You are connected to your SoundCloud account"));
        }
    }

    reply->deleteLater();
}

void SoundCloud::setCredentials(const QString &user, const QString &token) {
    this->setUsername(user);
    this->setAccessToken(token);
}

void SoundCloud::setUsername(const QString &user) {
    m_username = user;
    emit usernameChanged();
}

void SoundCloud::setAccessToken(const QString &token) {
    m_accessToken = token;
    emit accessTokenChanged();
    emit userSignedInChanged(this->userSignedIn());
}

void SoundCloud::postRequest(const QUrl &url, const QByteArray &data) {
    QNetworkRequest request(url);
    request.setHeader(QNetworkRequest::ContentTypeHeader, "application/x-www-form-urlencoded");
    request.setRawHeader("User-Agent", QString("MusiKloud/%1 (Qt)").arg(Utils::versionNumberString()).toUtf8());
    QNetworkReply *reply = this->networkAccessManager()->post(request, data);
    this->connect(reply, SIGNAL(finished()), this, SLOT(postFinished()));
    this->connect(this, SIGNAL(currentOperationCancelled()), reply, SLOT(deleteLater()));
}

void SoundCloud::putRequest(const QUrl &url) {
    QNetworkRequest request(url);
    request.setHeader(QNetworkRequest::ContentTypeHeader, "application/x-www-form-urlencoded");
    request.setRawHeader("User-Agent", QString("MusiKloud/%1 (Qt)").arg(Utils::versionNumberString()).toUtf8());
    QNetworkReply *reply = this->networkAccessManager()->put(request, QByteArray());
    this->connect(reply, SIGNAL(finished()), this, SLOT(postFinished()));
    this->connect(this, SIGNAL(currentOperationCancelled()), reply, SLOT(deleteLater()));
}

void SoundCloud::deleteRequest(const QUrl &url) {
    QNetworkRequest request(url);
    QNetworkReply *reply = this->networkAccessManager()->deleteResource(request);
    this->connect(reply, SIGNAL(finished()), this, SLOT(postFinished()));
    this->connect(this, SIGNAL(currentOperationCancelled()), reply, SLOT(deleteLater()));
}

void SoundCloud::postFinished() {
    QNetworkReply* reply = qobject_cast<QNetworkReply*>(this->sender());
    if (!reply) {
        emit error("Network error");
        return;
    }

    int statusCode = reply->attribute(QNetworkRequest::HttpStatusCodeAttribute).toInt();

    if ((statusCode == 200) || (statusCode == 201)) {
        emit postSuccessful(QString(reply->readAll()));
    }
    else {
        emit postFailed(reply->attribute(QNetworkRequest::HttpReasonPhraseAttribute).toString());
    }

    this->disconnect(this, SIGNAL(postSuccessful(QString)), 0, 0);
    this->disconnect(this, SIGNAL(postFailed(QString)), 0, 0);
    reply->deleteLater();
}

void SoundCloud::addComment(const QString &id, const QString &comment, int timestamp) {
    QUrl url(QString("https://api.soundcloud.com/tracks/%1/comments.json").arg(id));
    url.addQueryItem("oauth_token", this->accessToken());
    QByteArray data("&comment[body]=" + comment.toUtf8().toPercentEncoding());

    if (timestamp) {
        data += "&comment[timestamp]=" + QByteArray::number(timestamp);
    }

    this->postRequest(url, data);
    this->connect(this, SIGNAL(postSuccessful(QString)), this, SLOT(onCommentAdded(QString)));
}

void SoundCloud::onCommentAdded(const QString &response) {
    bool ok;
    QVariantMap comment = Json::parse(response, ok).toMap();

    if (ok) {
        emit commentAdded(QSharedPointer<CommentItem>(new CommentItem(comment)));
    }
    else {
        emit error(tr("Comment response invalid"));
    }
}

void SoundCloud::addToFavourites(const QStringList &ids) {
    if (!ids.isEmpty()) {
        m_actionIds = ids;
        m_actionsProcessed = 0;
        this->setBusy(true, tr("Adding track(s) to favourites"));
        this->addToFavourites();
    }
}

void SoundCloud::addToFavourites() {
    if (!m_actionIds.isEmpty()) {
        QUrl url(QString("https://api.soundcloud.com/me/favorites/%1.json").arg(m_actionIds.first()));
        url.addQueryItem("oauth_token", this->accessToken());
        this->putRequest(url);
        this->connect(this, SIGNAL(postSuccessful(QString)), this, SLOT(onAddedToFavourites()));
        this->connect(this, SIGNAL(postFailed(QString)), this, SLOT(onTrackActionError(QString)));
    }
}

void SoundCloud::onAddedToFavourites() {
    m_actionsProcessed++;
    emit busyProgressChanged(m_actionsProcessed);

    if (!m_actionIds.isEmpty()) {
        emit favouriteChanged(m_actionIds.takeFirst(), true);
    }
    if (!m_actionIds.isEmpty()) {
        this->addToFavourites();
    }
    else {
        this->setBusy(false);
        emit alert(tr("Track(s) added to favourites"));
    }
}

void SoundCloud::deleteFromFavourites(const QStringList &ids) {
    if (!ids.isEmpty()) {
        m_actionIds = ids;
        m_actionsProcessed = 0;
        this->setBusy(true, tr("Deleting track(s) from favourites"));
        this->deleteFromFavourites();
    }
}

void SoundCloud::deleteFromFavourites() {
    if (!m_actionIds.isEmpty()) {
        QUrl url(QString("https://api.soundcloud.com/me/favorites/%1.json").arg(m_actionIds.first()));
        url.addQueryItem("oauth_token", this->accessToken());
        this->deleteRequest(url);
        this->connect(this, SIGNAL(postSuccessful(QString)), this, SLOT(onDeletedFromFavourites()));
        this->connect(this, SIGNAL(postFailed(QString)), this, SLOT(onTrackActionError(QString)));
    }
}

void SoundCloud::onDeletedFromFavourites() {
    m_actionsProcessed++;
    emit busyProgressChanged(m_actionsProcessed);

    if (!m_actionIds.isEmpty()) {
        emit favouriteChanged(m_actionIds.takeFirst(), false);
    }
    if (!m_actionIds.isEmpty()) {
        this->deleteFromFavourites();
    }
    else {
        this->setBusy(false);
        emit alert(tr("Track(s) deleted from favourites"));
    }
}

void SoundCloud::onTrackActionError(const QString &errorString) {
    this->setBusy(false);
    emit error(errorString);
    m_actionIds.clear();
}

void SoundCloud::followUser(const QString &id) {
    m_actionId = id;
    QUrl url(QString("https://api.soundcloud.com/me/followings/%1.json").arg(id));
    url.addQueryItem("oauth_token", this->accessToken());
    this->putRequest(url);
    this->connect(this, SIGNAL(postSuccessful(QString)), this, SLOT(onUserFollowed()));
    this->connect(this, SIGNAL(postFailed(QString)), this, SLOT(onUserActionError(QString)));
}

void SoundCloud::onUserFollowed() {
    emit followingChanged(m_actionId, true);
    emit alert(tr("You have followed this user"));
}

void SoundCloud::unfollowUser(const QString &id) {
    m_actionId = id;
    QUrl url(QString("https://api.soundcloud.com/me/followings/%1.json").arg(id));
    url.addQueryItem("oauth_token", this->accessToken());
    this->deleteRequest(url);
    this->connect(this, SIGNAL(postSuccessful(QString)), this, SLOT(onUserUnfollowed()));
    this->connect(this, SIGNAL(postFailed(QString)), this, SLOT(onUserActionError(QString)));
}

void SoundCloud::onUserUnfollowed() {
    emit followingChanged(m_actionId, false);
    emit alert(tr("You no longer follow this user"));
}

void SoundCloud::onUserActionError(const QString &errorString) {
    emit error(errorString);
}

void SoundCloud::updateProfile(const QVariantMap &profile) {
    QUrl url("https://api.soundcloud.com/me.json");
    url.addQueryItem("oauth_token", this->accessToken());
    url.addQueryItem("user[full_name]", profile.value("fullname").toString());
    url.addQueryItem("user[discogs_name]", profile.value("bandname").toString());
    url.addQueryItem("user[website]", profile.value("websiteUrl").toString());
    url.addQueryItem("user[website_name]", profile.value("websiteTitle").toString());
    url.addQueryItem("user[description]", profile.value("description").toString());
    this->putRequest(url);
    this->connect(this, SIGNAL(postSuccessful(QString)), this, SLOT(onProfileUpdated(QString)));
    this->connect(this, SIGNAL(postFailed(QString)), this, SLOT(onUserActionError(QString)));
}

void SoundCloud::onProfileUpdated(const QString &response) {
    bool ok;
    QVariantMap profile = Json::parse(response, ok).toMap();

    if (ok) {
        emit profileUpdated(profile.value("id").toString(), profile);
        emit alert(tr("Your profile has been updated"));
    }
    else {
        emit error(tr("Cannot obtain new profile"));
    }
}

void SoundCloud::shareTrack(const QString &id, const QStringList &connections, const QString &message) {
    QUrl url(QString("https://api.soundcloud.com/tracks/%1/shared-to/connections").arg(id));
    url.addQueryItem("oauth_token", this->accessToken());
    QByteArray data(
                "&connections[][id]=" + connections.join("[id]=").toUtf8()
                + "&sharing_note=" + message.toUtf8().toPercentEncoding(QByteArray("+"))
                );
    this->postRequest(url, data);
    this->connect(this, SIGNAL(postSuccessful(QString)), this, SLOT(onTrackShared()));
}

void SoundCloud::onTrackShared() {
    emit alert(tr("Track shared to your connections"));
}
